class BinaryClassifiersAPI:
    def get_classifiers(self) -> list[str]:
        raise NotImplementedError

    def evaluate(self, *, classifier: str | None = None, text: str) -> bool:
        raise NotImplementedError



class TranslatorAPI:
    def get_lang_codes(self) -> list[str]:
        raise NotImplementedError

    def translate(self, *, text: str, lang_code_from: str | None = None, lang_code_to: str) -> str:
        raise NotImplementedError



from enum import StrEnum, auto


class DoctorChoiceXMLTagEnum(StrEnum):
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
        return f"{name.lower()}"

    DIAGNOSTICS = auto()
    SUMMARIZATION = auto()
    MTRS = auto()


class MTRSXMLTagEnum(StrEnum):
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
        return f"mtrs_{name.lower()}"

    NAME = auto()
    LABEL = auto()
    DESC = auto()


class MTRSLabelEnum(StrEnum):
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
        return name.upper()

    LABORATORY = auto()
    INSTRUMENTAL = auto()


class DiagnosticsXMLTagEnum(StrEnum):
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
        return f"diag_{name.lower()}"

    DIAG = auto()
    DOC = auto()
    DESC = auto()


class UncertaintyXMLTagEnum(StrEnum):
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list[str]) -> str:
        return f"{name.lower()}"

    LAST_MESSAGE = auto()
    INPUT_PROMPT = auto()
    RAW_DIAGNOSES = auto()
    U_ESTIMATE = auto()


from mmar_mapi.file_storage import FileStorageAPI, FileStorageBasic, FileStorage, ResourceId
from mmar_mapi.models.base import Base
from mmar_mapi.models.chat import Chat, Context, ChatMessage, AIMessage, HumanMessage, MiscMessage, make_content, Content, BaseMessage
from mmar_mapi.models.enums import MTRSLabelEnum, DiagnosticsXMLTagEnum, MTRSXMLTagEnum, DoctorChoiceXMLTagEnum, UncertaintyXMLTagEnum
from mmar_mapi.models.tracks import TrackInfo, DomainInfo
from mmar_mapi.models.widget import Widget
from mmar_mapi.utils import make_session_id, chunked
from mmar_mapi.xml_parser import XMLParser
from mmar_mapi.utils_import import load_main_objects
from mmar_mapi.decorators_maybe_lru_cache import maybe_lru_cache

__all__ = [
    "AIMessage",
    "Base",
    "BaseMessage",
    "Chat",
    "ChatMessage",
    "Content",
    "Context",
    "DiagnosticsXMLTagEnum",
    "DoctorChoiceXMLTagEnum",
    "UncertaintyXMLTagEnum",
    "DomainInfo",
    "FileStorage",
    "FileStorageAPI",
    "HumanMessage",
    "MTRSLabelEnum",
    "MTRSXMLTagEnum",
    "MiscMessage",
    "ResourceId",
    "TrackInfo",
    "Widget",
    "XMLParser",
    "chunked",
    "load_main_objects",
    "make_content",
    "make_session_id",
    "maybe_lru_cache",
]



from pydantic import BaseModel

from mmar_mapi.models.chat import Chat


Interpretation = str
ResourceId = str


class ContentInterpreterRemoteResponse(BaseModel):
    interpretation: str
    resource_fname: str
    resource: bytes


class ContentInterpreterRemoteAPI:
    def interpret_remote(
        self, *, kind: str, query: str, resource: bytes, chat: Chat | None = None
    ) -> ContentInterpreterRemoteResponse:
        raise NotImplementedError


class ContentInterpreterAPI:
    def interpret(
        self, *, kind: str, query: str, resource_id: str = "", chat: Chat | None = None
    ) -> tuple[Interpretation, ResourceId | None]:
        raise NotImplementedError



import json
from typing import Any

from pydantic import BaseModel, ConfigDict, model_validator


class Base(BaseModel):
    model_config = ConfigDict(str_strip_whitespace=True)

    @model_validator(mode="before")
    @classmethod
    def validate_to_json(cls, value: str | Any) -> Any:
        if isinstance(value, str):
            return cls(**json.loads(value))
        return value



from mmar_mapi.services.binary_classifiers import BinaryClassifiersAPI
from mmar_mapi.services.chat_manager import ChatManagerAPI
from mmar_mapi.services.content_interpreter import (
    ContentInterpreterAPI,
    ContentInterpreterRemoteAPI,
    ContentInterpreterRemoteResponse,
    Interpretation,
    ResourceId,
)
from mmar_mapi.services.critic import CriticAPI
from mmar_mapi.services.document_extractor import (
    DOC_SPEC_DEFAULT,
    DocExtractionOutput,
    DocExtractionSpec,
    DocumentExtractorAPI,
    ExtractedImage,
    ExtractedImageMetadata,
    ExtractedMarkdown,
    ExtractedPageImage,
    ExtractedPicture,
    ExtractedTable,
    ExtractionEngineSpec,
    ForceOCR,
    OutputType,
    PageRange,
)
from mmar_mapi.services.llm_hub import (
    LCP,
    RESPONSE_EMPTY,
    Attachments,
    LLMAccessorAPI,
    LLMCallProps,
    LLMEndpointMetadata,
    LLMHubAPI,
    LLMHubMetadata,
    LLMPayload,
    LLMRequest,
    LLMResponseExt,
    Message,
    Messages,
)
from mmar_mapi.services.text_extractor import TextExtractorAPI
from mmar_mapi.services.text_generator import TextGeneratorAPI
from mmar_mapi.services.text_processor import TextProcessorAPI
from mmar_mapi.services.translator import TranslatorAPI

__imported__ = [
    # LLM Hub
    LLMCallProps,
    LCP,
    Attachments,
    Message,
    Messages,
    LLMPayload,
    LLMRequest,
    LLMResponseExt,
    RESPONSE_EMPTY,
    LLMAccessorAPI,
    LLMHubAPI,
    LLMEndpointMetadata,
    LLMHubMetadata,
    # Document Extractor
    PageRange,
    ForceOCR,
    OutputType,
    ExtractionEngineSpec,
    DocExtractionSpec,
    ExtractedImage,
    ExtractedImageMetadata,
    ExtractedPicture,
    ExtractedTable,
    ExtractedPageImage,
    ExtractedMarkdown,
    DocExtractionOutput,
    DocumentExtractorAPI,
    DOC_SPEC_DEFAULT,
    # Chat Manager
    ChatManagerAPI,
    # Text Generator
    TextGeneratorAPI,
    # Content Interpreter
    ContentInterpreterAPI,
    ContentInterpreterRemoteAPI,
    ContentInterpreterRemoteResponse,
    Interpretation,
    # Binary Classifiers
    BinaryClassifiersAPI,
    # Translator
    TranslatorAPI,
    # Critic
    CriticAPI,
    # Text Processor
    TextProcessorAPI,
    # Text Extractor
    TextExtractorAPI,
    ResourceId,
]



from collections.abc import Callable
from functools import lru_cache

from loguru import logger


def maybe_lru_cache(maxsize: int, func: Callable) -> Callable:
    if maxsize >= 0:
        effective_maxsize: int | None = maxsize or None
        logger.info(f"Caching for {func.__name__}: enabled: maxsize={effective_maxsize}")
        func = lru_cache(maxsize=effective_maxsize)(func)
    else:
        logger.info(f"Caching for {func.__name__}: disabled")
    return func



from abc import ABC, ABCMeta, abstractmethod

from loguru import logger

from mmar_mapi.models.chat import AIMessage, Chat, ChatMessage, Content, HumanMessage
from mmar_mapi.utils_import import load_main_objects

State = str
TrackResponse = AIMessage | list[AIMessage] | tuple[State, Content]
DOMAIN_CAPTION = {"DOMAIN", "CAPTION"}


def check_class_has_attrs(name, classdict, attrs):
    for attr in attrs:
        if attr not in classdict:
            raise TypeError(f"Class '{name}' must define a '{attr}'")


class MustHaveDomainCaptionMeta(ABCMeta):
    def __new__(cls, name, bases, classdict):
        is_abstract = any({getattr(v, "__isabstractmethod__", None) for v in classdict.values()})

        # Check if "KIND" is defined in the class being created
        if not is_abstract:
            check_class_has_attrs(name, classdict, DOMAIN_CAPTION)
        return super().__new__(cls, name, bases, classdict)


class TrackI(ABC, metaclass=MustHaveDomainCaptionMeta):
    @abstractmethod
    def get_response(self, chat: Chat) -> list[ChatMessage]:
        pass


class StateActionPolicyTrack(TrackI):
    def get_response(self, chat: Chat) -> list[ChatMessage]:
        user_message = chat.get_last_user_message()
        if not user_message:
            # assuming here that track is reactive: if last message is not from user, no reaction needed
            logger.warning("Not user message found! Returning empty")
            return []

        state = self.select_state(chat, user_message)
        action = self.select_action(chat, user_message, state=state)
        content = self.generate_response_content(chat, user_message, action=action)

        logger.debug(f"Bot response: '{content}'")
        extra = dict(action=action)
        bot_message = AIMessage(content=content, extra=extra, state=state)

        messages = [bot_message]
        bot_message = messages[-1]
        prev_state = chat.get_last_state()
        cur_state = bot_message.state
        # todo implement moderation
        logger.info(f"State transition: '{prev_state}' -> '{cur_state}', action: '{action}'")
        return messages  # type: ignore[return-value]

    @abstractmethod
    def select_state(self, chat: Chat, user_message: HumanMessage) -> str:
        pass

    @abstractmethod
    def select_action(self, chat: Chat, user_message: HumanMessage, state: str) -> str:
        pass

    @abstractmethod
    def generate_response_content(self, chat: Chat, user_message: HumanMessage, action: str) -> Content:
        pass


class SimpleTrack(TrackI):
    def get_response(self, chat: Chat) -> list[ChatMessage]:
        user_message = chat.get_last_user_message()
        if not user_message:
            # assuming here that track is reactive: if last message is not from user, no reaction needed
            logger.warning("Not user message found! Returning empty")
            return []

        response: TrackResponse = self.generate_response(chat, user_message)
        if isinstance(response, tuple) and len(response) == 2:
            state: str = response[0]
            content: Content = response[1]
            messages = [AIMessage(state=state, content=content, extra=None)]
        elif isinstance(response, str):
            state, content = "", response
            messages = [AIMessage(state=state, content=content, extra=None)]
        elif isinstance(response, AIMessage):
            state = response.state
            messages = [response]
        elif isinstance(response, list):
            if response:
                last_msg = response[-1]
                state = last_msg.state if last_msg.is_ai else ""
                messages = response
            else:
                state = ""
                messages = []
        else:
            logger.error(f"Bad response type: {type(response)}: {response}")
            state = ""
            messages = []
        logger.debug(f"Bot response: '{messages}'")

        track_id = chat.context.track_id
        logger.info(f"Track {track_id}: state transition: '{chat.get_last_state()}' -> '{state}'")
        return messages  # type: ignore[return-value]

    @abstractmethod
    def generate_response(self, chat: Chat, user_message: HumanMessage) -> TrackResponse:
        pass


def load_tracks(tracks_module) -> dict[str, TrackI]:
    return load_main_objects(tracks_module, TrackI)  # type: ignore[return-value]



from typing import Literal, cast

from pydantic import BaseModel, ConfigDict

from mmar_mapi import ChatMessage, BaseMessage


class LLMCallProps(BaseModel, frozen=True):
    model_config = ConfigDict(extra="forbid")

    endpoint_key: str = ""
    attempts: int = 1
    extra: dict[str, str] | None = None

    def with_endpoint_key(self, endpoint_key) -> 'LLMCallProps':
        return self.model_copy(update=dict(endpoint_key=endpoint_key))

    def put_extra(self, extra_key: str, extra_value: str) -> 'LLMCallProps':
        extra_upd = {**(self.extra or {}), extra_key: extra_value}
        return self.model_copy(update=dict(extra=extra_upd))
        

LCP = LLMCallProps()
ResourceId = str
FileId = str
Attachments = list[list[ResourceId]]
Role = Literal["system", "assistant", "user"]


class Message(BaseModel, frozen=True):
    role: Role
    content: str

    @staticmethod
    def create(message: ChatMessage) -> "Message | None":
        return _create_message(message=message)

    def get_content(self):
        return self.content


def _create_message(message: ChatMessage) -> Message | None:
    msg = cast(BaseMessage, message)
    role: Role | None = "assistant" if msg.is_ai else "user" if msg.is_human else None
    return Message(role=role, content=msg.text) if role else None


class Messages(BaseModel, frozen=True):
    messages: list[Message]


class LLMPayload(Messages, frozen=True):
    attachments: Attachments | None = None

    def with_attachments(self, attachments: Attachments) -> "LLMPayload":
        return self.model_copy(update=dict(attachments=attachments))

    def __repr__(self):
        return self.show_pretty()

    def show_pretty(self, detailed: bool=False):
        total_size = sum(len(msg.content) for msg in self.messages)
        parts = [
            f"messages: {len(self.messages)}",
            f"total size: {total_size}" if detailed else None,
            "has attachments" if self.attachments else None,
        ]
        payload_pretty = ", ".join(filter(None, parts))
        return f"LLMPayload({payload_pretty})"

    @staticmethod
    def create(user_text: str, resource_id: ResourceId = "") -> "LLMPayload":
        return _create_payload(user_text=user_text, resource_id=resource_id)

    @staticmethod
    def parse(request: "Request") -> "LLMPayload":
        return _parse_payload(request)

    def get_resource_id(self) -> str | None:
        if not self.attachments:
            return None
        resource_id = self.attachments[0][0]
        return resource_id



def _create_payload(user_text: str, resource_id: ResourceId = ""):
    payload = LLMPayload(messages=[Message(role="user", content=user_text)])
    if not resource_id:
        return payload
    else:
        return payload.with_attachments(attachments=[[resource_id]])


class LLMResponseExt(BaseModel):
    text: str
    resource_id: ResourceId | None = None


RESPONSE_EMPTY = LLMResponseExt(text="")
LLMRequest = str | list[Message] | LLMPayload

def _parse_payload(request: LLMRequest) -> LLMPayload:
    if isinstance(request, str):
        return LLMPayload(messages=[Message(role="user", content=request)])
    elif isinstance(request, list) and all(isinstance(msg, Message) for msg in request):
        return LLMPayload(messages=request)
    elif isinstance(request, list) and all(isinstance(msg, dict) for msg in request):
        # todo validate contents of `msg`
        return LLMPayload(messages=request)
    elif isinstance(request, LLMPayload):
        return request
    elif isinstance(request, dict):
        # todo validate request fields
        return LLMPayload.model_validate(request)
    else:
        raise ValueError(f"Bad request type {type(request)}: {request}")


class LLMEndpointMetadata(BaseModel):
    key: str
    caption: str


class LLMHubMetadata(BaseModel):
    endpoints: list[LLMEndpointMetadata]
    default_endpoint_key: str

    def get_endpoint_keys(self):
        return [ep.key for ep in self.endpoints]

LLM_HUB_METADATA_EMPTY = LLMHubMetadata(endpoints=[], default_endpoint_key='')

class LLMHubAPI:
    def get_metadata(self) -> LLMHubMetadata:
        raise NotImplementedError

    def get_response(self, *, request: LLMRequest, props: LLMCallProps = LCP) -> str:
        raise NotImplementedError

    def get_response_ext(self, *, request: LLMRequest, props: LLMCallProps = LCP) -> LLMResponseExt:
        raise NotImplementedError

    def get_embedding(self, *, prompt: str, props: LLMCallProps = LCP) -> list[float] | None:
        raise NotImplementedError

# will be removed in the future
Request = LLMRequest
ResponseExt = LLMResponseExt
LLMAccessorAPI = LLMHubAPI



from enum import StrEnum
from typing import Annotated

from pydantic import AfterValidator, BaseModel, model_validator


def _validate_page_range(v: tuple[int, int]) -> tuple[int, int]:
    if v[0] < 1 or v[1] < v[0]:
        raise ValueError("Invalid page range: start must be ≥ 1 and end must be ≥ start.")
    return v


PageRange = Annotated[tuple[int, int], AfterValidator(_validate_page_range)]
ForceOCR = StrEnum("ForceOCR", ["ENABLED", "DISABLED", "AUTO"])
OutputType = StrEnum("OutputType", ["RAW", "PLAIN", "MARKDOWN"])
ResourceId = str


class ExtractionEngineSpec(BaseModel, frozen=True):
    output_type: OutputType = OutputType.MARKDOWN
    force_ocr: ForceOCR = ForceOCR.AUTO
    do_ocr: bool = False
    do_table_structure: bool = False
    do_cell_matching: bool = False
    do_annotations: bool = False
    do_image_extraction: bool = False
    generate_page_images: bool = False
    images_scale: float = 2.0


class DocExtractionSpec(BaseModel, frozen=True):
    page_range: PageRange | None = None
    engine: ExtractionEngineSpec = ExtractionEngineSpec()

    def _update(self, **update):
        return self.model_copy(update=update)

    def _update_engine(self, **engine_update):
        return self._update(engine=self.engine.model_copy(update=engine_update))

    # fmt: off
    def with_output_type_raw(self): return self._update_engine(output_type=OutputType.RAW)
    def with_output_type_plain(self): return self._update_engine(output_type=OutputType.PLAIN)
    def with_ocr(self): return self._update_engine(do_ocr=True)
    def with_tables(self): return self._update_engine(do_table_structure=True, do_cell_matching=True)
    def with_images(self): return self._update_engine(do_image_extraction=True)
    def with_annotations(self): return self._update_engine(do_annotations=True)
    def with_force_ocr_enabled(self): return self._update_engine(force_ocr=ForceOCR.ENABLED)
    def with_force_ocr_disabled(self): return self._update_engine(force_ocr=ForceOCR.DISABLED)
    def with_page_images(self): return self._update_engine(generate_page_images=True)

    def with_page_range(self, page_range: PageRange): return self._update(page_range=page_range)
    # fmt: on

DOC_SPEC_DEFAULT = DocExtractionSpec()


class ExtractedImage(BaseModel):
    page: int
    image_resource_id: ResourceId | None = None


class ExtractedImageMetadata(BaseModel):
    annotation: str = ""
    caption: str = ""
    width: int | None = None
    height: int | None = None


class ExtractedPicture(ExtractedImage, ExtractedImageMetadata):
    "Image of part of page"

    ...


class ExtractedTable(ExtractedImage, ExtractedImageMetadata):
    formatted_str: str


class ExtractedPageImage(ExtractedImage):
    "Image of all page"

    pass

class ExtractedMarkdown(BaseModel):
    page: int
    text: str


def _validate_no_duplicate_markdowns(v: list[ExtractedMarkdown]) -> list[ExtractedMarkdown]:
    """Validate that no two ExtractedMarkdown items have the same page number."""
    seen_pages = set()
    for item in v:
        if item.page in seen_pages:
            raise ValueError(f"Duplicate page number: {item.page}")
        seen_pages.add(item.page)
    return v


class DocExtractionOutput(BaseModel):
    spec: DocExtractionSpec
    text: str = "" # `text` is a concatenation of all `markdowns[*].text` (pages joined together)
    tables: list[ExtractedTable] = []
    pictures: list[ExtractedPicture] = []
    page_images: list[ExtractedPageImage] = []
    markdowns: Annotated[list[ExtractedMarkdown], AfterValidator(_validate_no_duplicate_markdowns)] = []

    @model_validator(mode="after")
    def _validate_pages_in_range(self) -> "DocExtractionOutput":
        """Validate that all page numbers are within spec.page_range if specified."""
        if self.spec.page_range is None:
            return self
        start, end = self.spec.page_range
        def _check(name: str, page: int) -> None:
            if page < start or page > end:
                raise ValueError(f"{name} page {page} is outside of specified range [{start}, {end}]")
        for item in self.markdowns:
            _check("markdowns", item.page)
        for item in self.tables:
            _check("tables", item.page)
        for item in self.pictures:
            _check("pictures", item.page)
        for item in self.page_images:
            _check("page_images", item.page)
        return self

class DocumentExtractorAPI:
    def extract(self, *, resource_id: ResourceId, spec: DocExtractionSpec = DOC_SPEC_DEFAULT) -> ResourceId | None:
        """returns file with DocExtractionOutput"""
        raise NotImplementedError



from mmar_mapi.models.chat import Chat, ChatMessage
from mmar_mapi.models.tracks import DomainInfo, TrackInfo


class ChatManagerAPI:
    def get_domains(self, *, client_id: str, language_code: str = "ru") -> list[DomainInfo]:
        raise NotImplementedError

    def get_tracks(self, *, client_id: str, language_code: str = "ru") -> list[TrackInfo]:
        raise NotImplementedError

    def get_response(self, *, chat: Chat) -> list[ChatMessage]:
        raise NotImplementedError
